#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "lib/log.h"
#include "lib/safem.h"
#include "texture.h"
#include "wad/wadstruct.h"

/* Uncomment this for loads of debugging msgs */
/*#define TXDEBUG*/

int texture_flags=0;

/* playpal stuff */

static LumpNum ppln;
static const unsigned char *playpal=NULL;

static void init_playpal(void) {
   if(playpal==NULL) {
      ppln=getlump("PLAYPAL");
      playpal=load_lump(ppln);
   };
};
static void reset_playpal(void) {
   if(playpal!=NULL) {
      free_lump(ppln);
      playpal=NULL;
   };
};

void set_playpal(int idx,SetPalFunc func) {
   int i=256;
   const unsigned char *p;
   init_playpal();
   p=playpal+(idx+1)*256*3;
   while(i>0) {
      i--;
      p-=3;
      func(i,p[0],p[1],p[2]);
   }
};

/* pixcvt stuff */

extern inline unsigned short pix8topix16(unsigned char p) { 
   const unsigned char *pp=playpal+(int)p*3;
   LE_int16 _r;
   unsigned short r;
   if(p==255) return 0xffff;
   _r=(pp[2]>>3)|((pp[1]>>2)<<5)|((pp[0]>>3)<<11);
   r=*(const short *)&_r;
   /* 0xffff means transparent, so never produce it to mean white */
   if(r==0xffff) r=0xfffe;
   return r;
};
extern inline unsigned int pix8topix32(unsigned char p) {
   const unsigned char *pp=playpal+(int)p*3;
   union {
      unsigned int r;
      struct {
	 char a,b,c;
      } i;
   } u;
   if(p==255) return 0xffffffff;
   u.r=0;
   u.i.a=pp[2];
   u.i.b=pp[1];
   u.i.a=pp[0];
   return u.r;
};

#define Pixel unsigned char
#define ppot ppot8
#define jpot jpot8
#define pixcvt(p) p
#include "pastepic.c"
#undef Pixel
#undef ppot 
#undef jpot 
#undef pixcvt

#define Pixel unsigned short
#define ppot ppot16
#define jpot jpot16
#define pixcvt pix8topix16
#include "pastepic.c"
#undef Pixel
#undef ppot 
#undef jpot 
#undef pixcvt

#define Pixel unsigned int
#define ppot ppot32
#define jpot jpot32
#define pixcvt pix8topix32
#include "pastepic.c"
#undef Pixel
#undef ppot 
#undef jpot 
#undef pixcvt

static void paste_pict_on_texture(Texture *t,const PictData *p,int x,int y) {
   switch(t->bpp) {
    case(1): ppot8(t,p,x,y); break;
    case(2): init_playpal(); ppot16(t,p,x,y); break;
    case(4): init_playpal(); ppot32(t,p,x,y); break;
    default:
      logprintf(LOG_ERROR,'T',"Bad BBP value in paste_pict");
   };
};
static void paste_jpatch_on_texture(Texture *t,const AltPictData *p,int x,int y) {
   switch(t->bpp) {
    case(1): jpot8(t,p,x,y); break;
    case(2): init_playpal(); jpot16(t,p,x,y); break;
    case(4): init_playpal(); jpot32(t,p,x,y); break;
    default:
      logprintf(LOG_ERROR,'T',"Bad BBP value in paste_jpatch");
   };
};

/*** PNAMES stuff ***/

/* *not* the same format as in the .WAD --- these are null-terminated */
static int npnames=0;
static char *pnames=NULL;

static void reset_pnames(void) {
   if(pnames) {
      npnames=0;
      safe_free(pnames);
      pnames=NULL;
   };
};
static void init_pnames(void) {
   LumpNum ln;
   const char *pn;
   int i;
   if(pnames!=NULL) return;
   ln=getlump("PNAMES");
   pn=load_lump(ln);
   npnames=*(const LE_int32 *)(pn);
   pn+=4;
   pnames=safe_calloc(npnames,9);
   for(i=0;i<npnames;i++)
     strncpy(pnames+9*i,pn+8*i,8);
   free_lump(ln);
   logprintf(LOG_DEBUG,'T',"npnames=%d (should be %d)",npnames,
	     (get_lump_len(ln)-4)/8);
}
#define lookup_pname(i) (pnames+9*i)

/*** TEXTURE1 stuff ***/

static int nwalltexs=0;
static LumpNum t1ln=BAD_LUMPNUM,t2ln=BAD_LUMPNUM;
static const TextureTable *texture1=NULL,*texture2=NULL;

static void init_texture_lumps(void) {
   if(texture1==NULL) {
      t1ln=lookup_lump("TEXTURE1",NULL,NULL);
      t2ln=lookup_lump("TEXTURE2",NULL,NULL);
      if(t1ln==BAD_LUMPNUM)
	logprintf(LOG_FATAL,'T',"No metatexture lumps in WAD");
      texture1=load_lump(t1ln);
      if(t2ln==BAD_LUMPNUM) texture2=NULL;
      else texture2=load_lump(t2ln);
      nwalltexs=texture1->UMEMB(hdr).ntxts;
      logprintf(LOG_DEBUG,'T',"%s has %d wall textures","TEXTURE1",
		(int)(texture1->UMEMB(hdr).ntxts));
      if(texture2) {
	 nwalltexs+=texture2->UMEMB(hdr).ntxts;
	 logprintf(LOG_DEBUG,'T',"%s has %d wall textures","TEXTURE2",
		   (int)(texture2->UMEMB(hdr).ntxts));
      };
   }
}
static void reset_texture_lumps(void) {
   if(t1ln!=BAD_LUMPNUM) free_lump(t1ln);
   if(t2ln!=BAD_LUMPNUM) free_lump(t2ln);
   t1ln=t2ln=BAD_LUMPNUM;
   texture1=texture2=NULL;
   nwalltexs=0;
}

static const TextureData *find_tdata(int i) {
   const TextureTable *tt=texture1;
   if(i>=texture1->UMEMB(hdr).ntxts) {
      tt=texture2;
      i-=texture1->UMEMB(hdr).ntxts;
   }
   return (const TextureData *)(tt->data+(tt->UMEMB(hdr).idx[i]));
};


/*** Some generic Texture helpers ***/

static int mylog2(int i) {
   int j=0;
   while(j<32&&i>(1<<j)) j++;
   return j;
};

static void free_many_textures(Texture *tt,int i) {
   while(i>0) free_texels(tt+(--i));
}

static Texture *find_texture(const char *name,Texture *tt,int n) {
   int i;
   for(i=0;i<n;i++) 
     if(!strncasecmp(tt[i].name,name,8)) return tt+i;
   return NULL;
}


/*** sprites ***/

static int nsprites=0,topsprite=0;
static Texture *sprites=NULL;

static void init_sprites(void) {
   if(sprites) return;
   nsprites=count_lumps_between("S_START","S_END");
   if(texture_flags&TF_RELAX_SPRITES)
     nsprites+=256; /* so, a relaxed sprite PWAD can't be bigger than this */
   if(nsprites<=0) logfatal('T',"Can't find any sprites in WAD");
   logprintf(LOG_INFO,'T',"Init %d sprites",nsprites);
   sprites=(Texture *)safe_calloc(nsprites,sizeof(Texture));
   topsprite=0;
};
static void reset_sprites(void) {
   if(sprites) {
      free_many_textures(sprites,nsprites);
      safe_free(sprites);
      sprites=NULL;
      nsprites=0;
   }
}

static void load_sprite(Texture *t,int bpp) {
   const PictData *pd;
   size_t l;
   if(!LUMPNUM_OK(t->lumpnum)) 
      logfatal('T',"Bad lumpnum for texture %s in load_sprite()",t->name);
   pd=load_lump(t->lumpnum);
   l=bpp<<(t->log2width+t->log2height);
   if(IS_JPATCH(pd)&&bpp==1) {
      t->bpp=1;
      t->texels=(void *)pd->UMEMB(alt).data;
      t->alloced_texels=0;
   }
   else {
      t->bpp=bpp;
      t->texels=safe_malloc(l);
      t->alloced_texels=1;
      memset(t->texels,-1,l);
      if(IS_JPATCH(pd)) 
	 paste_jpatch_on_texture(t,&(pd->UMEMB(alt)),0,0);
      else 
	 paste_pict_on_texture(t,pd,0,0);
      free_lump(t->lumpnum);
   };
};

static void free_sprite(Texture *t) {
   if(t->alloced_texels) safe_free(t->texels);
   else free_lump(t->lumpnum);
   t->texels=NULL;
   t->alloced_texels=0;
   t->bpp=0;
};

static void guess_sprite_size(Texture *t) {
   const PictData *pd=load_lump(t->lumpnum);
   if(IS_JPATCH(pd)) {
      t->width=pd->UMEMB(alt).width;
      t->height=pd->UMEMB(alt).height;
      t->log2width=pd->UMEMB(alt).log2width;
      t->log2height=pd->UMEMB(alt).log2height;
   }
   else {
      t->width=pd->UMEMB(hdr).width;
      t->height=pd->UMEMB(hdr).height;
      t->log2width=mylog2(t->width);
      t->log2height=mylog2(t->height);
   };
   release_lump(t->lumpnum);
};

Texture *get_sprite_texture(const char *name) {
   Texture *t;
   init_sprites();
   t=find_texture(name,sprites,nsprites);
   if(t) return t;
   if(topsprite>=nsprites) logprintf(LOG_FATAL,'T',"Too many sprites???");
   t=sprites+topsprite++;
   t->type=TT_SPRITE;
   if(texture_flags&TF_RELAX_SPRITES)
     t->lumpnum=getlump(name);
   else
     t->lumpnum=safe_lookup_lump(name,"S_START","S_END",LOG_FATAL);
   strncpy(t->name,name,8);
   t->opaque=0;
   guess_sprite_size(t);
   return t;
};

Texture *get_sprite(const char *sp,char ph,char r)  {
   LumpNum ln;
   char buf[10],buf2[10];
   buf[8]=0;
   memcpy(buf,sp,4);
   buf[4]=ph;
   buf[5]='0';
   buf[6]='?';
   buf[7]='?';
   if(texture_flags&TF_RELAX_SPRITES)
     ln=lookup_lump(buf,NULL,NULL);
   else ln=lookup_lump(buf,"S_START","S_END");
   if(LUMPNUM_OK(ln)) return get_sprite_texture(get_lump_name(ln,buf2));
   buf[4]=ph;
   buf[5]=r;
   buf[6]='?';
   buf[7]='?';
   if(texture_flags&TF_RELAX_SPRITES)
     ln=lookup_lump(buf,NULL,NULL);
   else ln=lookup_lump(buf,"S_START","S_END");
   if(LUMPNUM_OK(ln)) return get_sprite_texture(get_lump_name(ln,buf2));
   buf[6]=ph;
   buf[7]=r;
   buf[4]='?';
   buf[5]='?';
   if(texture_flags&TF_RELAX_SPRITES)
     ln=lookup_lump(buf,NULL,NULL);
   else ln=lookup_lump(buf,"S_START","S_END");
   if(LUMPNUM_OK(ln)) return get_sprite_texture(get_lump_name(ln,buf2));
   return NULL;
};

/*** flats ***/

static int nflats=0;
static Texture *flats=NULL;

static void init_flats(void) {
   LumpNum ln=0;
   int topflat=0;
   if(flats) return;
   nflats=count_lumps_between("F_START","F_END");
   if(nflats<=0) logprintf(LOG_FATAL,'T',"Can't find any flats in WAD");
   flats=(Texture *)safe_calloc(nflats,sizeof(Texture));
   while(topflat<nflats) {
      ln=lumplook(ln,"F_START");
      if(!LUMPNUM_OK(ln)) break;
      while(topflat<nflats) {
	 int i;
	 Texture *t=flats+topflat;
	 t->lumpnum=ln=lumpnext(ln,0);
	 if(!LUMPNUM_OK(ln)) break;
	 get_lump_name(ln,t->name);
	 if(!strcasecmp(t->name,"F_END")) break;
	 t->type=TT_FLAT;
	 t->opaque=1;
	 switch(i=get_lump_len(t->lumpnum)) {
	 case(0): break;
	 case(128*128):
	    t->width=t->height=128;
	    t->log2width=t->log2height=7;
	    topflat++;
	    break;
	 case(64*64):
	    t->width=t->height=64;
	    t->log2width=t->log2height=6;
	    topflat++;
	    break;
	 default:
	    logfatal('T',"funny length (%d) for flat %s",i,t->name);
	 };
      };
   };
   logprintf(LOG_INFO,'T',"Init %d flats",nflats=topflat);
};
static void reset_flats(void) {
   if(flats) {
      free_many_textures(flats,nflats);
      safe_free(flats);
      flats=NULL;
      nflats=0;
   }
}

static void load_flat(Texture *t,int bpp) {
   t->bpp=bpp;
   if(bpp==1) t->texels=(void *)load_lump(t->lumpnum);
   else {
      int i=t->width*t->height;
      const unsigned char *data=load_lump(t->lumpnum);
      t->texels=safe_malloc(i*bpp);
      while(i>0) {
	 i--;
	 if(bpp==2) ((unsigned short *)(t->texels))[i]=pix8topix16(data[i]);
	 else if(bpp==4) ((unsigned int *)(t->texels))[i]=pix8topix32(data[i]);
      }
      free_lump(t->lumpnum);
   }
}

static void free_flat(Texture *t) {
   if(t->texels) {
      if(t->bpp==1) free_lump(t->lumpnum);
      else safe_free(t->texels);
   }
   t->texels=NULL;
   t->bpp=0;
}

Texture *get_flat_texture(const char *name) {
   Texture *t;
   init_flats();
   t=find_texture(name,flats,nflats);
   if(t==NULL) 
     logfatal('T',"Can't find flat texture %s",name);
   return t;
}

/*** patches ***/

static LumpNum *patlns=NULL;
static const PictData **patlumps=NULL;

static void init_patches(void) {
   int i;
   if(patlns) return;
   init_pnames();
   patlns=safe_calloc(npnames,sizeof(LumpNum));
   patlumps=safe_calloc(npnames,sizeof(const PictHeader *));
   for(i=0;i<npnames;i++)
     patlns[i]=lookup_lump(lookup_pname(i),NULL,NULL);
};
static void reset_patches(void) {
   int i;
   /* 
      NB: though it is OK to free id-style patches at any time,
      J1 patches need to stay resident, as they may be the sole
      patch of a 1bpp texture, in which case the raw patch data will
      be referenced by the texture.
      Earlier versions of texture.c had this function as a non-static
      so that patches could be freed as part of a periodic garbage
      collection.  This is no longer safe.
   */
   if(!patlns) return;
   for(i=0;i<npnames;i++) 
     if(LUMPNUM_OK(patlns[i])) free_lump(patlns[i]);
   safe_free(patlns);
   safe_free(patlumps);
   patlns=NULL;
   patlumps=NULL;
   reset_pnames();
};

static const PictData *get_patch_data(int num) {
   if(patlumps[num]==NULL)  {
      if(!LUMPNUM_OK(patlns[num])) 
	 logprintf(LOG_FATAL,'T',"Patchnum %d (%s) doesn't exist"
		   ,num,pnames+num*9);
      else {
#ifdef TXDEBUG
	 logprintf(LOG_DEBUG,'T',"loading patchnum %d (%s)"
		   ,num,pnames+num*9);
#endif
	 patlumps[num]=load_lump(patlns[num]);
      };
   };
   return patlumps[num];
};


/*** walls ***/

static Texture *walltexs=NULL;

static void init_walltexs(void) {
   int i;
   init_texture_lumps();
   init_patches();
   if(walltexs) return;
#ifdef TXDEBUG
   logprintf(LOG_DEBUG,'T',"init_walltexs(): %d textures",nwalltexs);
#endif
   walltexs=(Texture *)safe_calloc(sizeof(Texture),nwalltexs);
   for(i=0;i<nwalltexs;i++) {
      Texture *t=walltexs+i;
      const TextureData *td=find_tdata(i);
      strncpy(t->name,td->name,8);
      t->type=TT_WALL;
      t->opaque=1; /* a lie, but good enough for now */
      t->alloced_texels=0;
      t->width=td->dx;
      t->height=td->dy;
      t->log2width=mylog2(td->dx);
      t->log2height=mylog2(td->dy);
   };
};
static void reset_walltexs(void) {
   if(walltexs) {
      free_many_textures(walltexs,nwalltexs);
      safe_free(walltexs);
   };
};

Texture *get_wall_texture(const char *name) {
   Texture *t;
   if(*name=='&') return get_flat_texture(name+1);
   init_walltexs();
   t=find_texture(name,walltexs,nwalltexs);
   if(t==NULL) 
     logprintf(LOG_FATAL,'T',"Can't find wall texture %s",name);
   return t;
};

static void load_wtex(Texture *t,int bpp) {
   const TextureData *td;
   int i;
   init_walltexs();
   t->bpp=bpp;
   td=find_tdata(t-walltexs);
#ifndef NO_1P_KLUDGE
   /* if this is a one-patch texture, try this kludge */
   /* apparently something is wrong with this code, causing it to */
   /* crash on DEC Alphas.  I'm not sure what it is (josh) */
   if(bpp==1&&td->npatches==1&&td->patch[0].x==0&&td->patch[0].y==0) {
      const PictData *p=get_patch_data(td->patch[0].pnum);
      if(IS_JPATCH(p)&&p->UMEMB(alt).log2height==t->log2height) {
	 t->texels=(void *)p->UMEMB(alt).data;
#ifdef TXDEBUG
	 logprintf(LOG_DEBUG,'T',"using one-patch kludge for %s",td->name);
#endif
	 return;
      };
   };
#endif   
#ifdef TXDEBUG
   logprintf(LOG_DEBUG,'T',"load_walltex %s at %d bpp",t->name,bpp);
   logprintf(LOG_DEBUG,'T',"l2w=%d l2h=%d",t->log2width,t->log2height);
#endif
   t->texels=safe_malloc(i=(bpp<<(t->log2width+t->log2height)));
   t->alloced_texels=1;
   memset(t->texels,-1,i);
#ifdef TXDEBUG
   logprintf(LOG_DEBUG,'T',"%s: %d patches",t->name,(int)(td->npatches));
#endif
   for(i=0;i<td->npatches;i++) {
      const PictData *p=get_patch_data(td->patch[i].pnum);
      if(IS_JPATCH(p))
	 paste_jpatch_on_texture(t,&(p->UMEMB(alt)),
				 td->patch[i].x,td->patch[i].y);
      else if(p->UMEMB(hdr).height<255)
	 paste_pict_on_texture(t,p,td->patch[i].x,td->patch[i].y);
      else 
	 logfatal('T',"Can't make sense of patch #%d, texture %s",
		  (int)(td->patch[i].pnum),td->name);
   };
#ifdef TXDEBUG
   logprintf(LOG_DEBUG,'T',"%s: loaded",t->name);
#endif
};

static void free_wtex(Texture *t) {
   if(t->alloced_texels) safe_free(t->texels);
   t->texels=NULL;
   t->alloced_texels=0;
   t->bpp=0;
};


/*** FONTS ***/

#define MAXFONTS 16
#define MAXFONTNAME 16

static Texture *fonts[MAXFONTS];
static char fontname[MAXFONTS*MAXFONTNAME];
static int nfontents[MAXFONTS],nfonts=0;

int init_font(const char *format,int nchars) {
   int i;
   for(i=0;i<nfonts;i++)
      if(!strncmp(fontname+i*MAXFONTNAME,format,MAXFONTNAME)) {
	 logprintf(LOG_DEBUG,'T',"found font %s already loaded",format);
	 return i;
      };
   if(nfonts>=MAXFONTS) logfatal('T',"Too many fonts");
   fonts[nfonts]=(Texture *)safe_calloc(nfontents[nfonts]=nchars,
					sizeof(Texture));
   strncpy(fontname+nfonts*MAXFONTNAME,format,MAXFONTNAME);
   /* OK, now set up each character to be loaded when needed */
   logprintf(LOG_INFO,'T',"Loading font #%d (%s)",nfonts,format);
   for(i=0;i<nchars;i++) {
      Texture *t=fonts[nfonts]+i;
      t->opaque=0;
      t->type=TT_FONT;
      sprintf(t->name,format,i);
      t->lumpnum=lookup_lump(t->name,NULL,NULL);
      if(LUMPNUM_OK(t->lumpnum)) guess_sprite_size(t);
   };
   /* done */
   return nfonts++;
};

static void reset_fonts(void) {
   while(nfonts>0) {
      nfonts--;
      free_many_textures(fonts[nfonts],nfontents[nfonts]);
      safe_free(fonts[nfonts]);
   };
};

Texture *get_font_texture(int fontnum,unsigned char ch) {
   if(fontnum<0||fontnum>=nfonts) 
      logfatal('T',"Request for bad fontnum (%d)",fontnum);
   if(ch>=nfontents[fontnum]) 
      logfatal('T',"Request for bad font char (%d:%d)",fontnum,ch);
   return fonts[fontnum]+ch;
};


/*** MISCELLANEOUS TEXTURES ***/

/* since misc. textures aren't used for much, this should be OK */
#define MAXMISC 32
static Texture misc[MAXMISC];
int nmisc=0;

Texture *get_misc_texture(const char *name) {
   Texture *t=find_texture(name,misc,nmisc);
   if(t) return t;
   if(nmisc>=MAXMISC) 
      logfatal('T',"too many misc textures");
   t=misc+nmisc++;
   t->type=TT_MISC;
   t->opaque=0;
   strncpy(t->name,name,8);
   t->lumpnum=safe_lookup_lump(name,NULL,NULL,LOG_FATAL);
   guess_sprite_size(t);
   return t;
};

static void reset_misc(void) {
   free_many_textures(misc,nmisc);
};

/*** OTHER STUFF ***/

int get_texture_num(TextureType tt,const char *name) {
   switch(tt) {
   case(TT_FLAT): return find_texture(name,flats,nflats)-flats; break;
   case(TT_WALL): return find_texture(name,walltexs,nwalltexs)-walltexs; break;
   default: logprintf(LOG_ERROR,'T',
		      "get_texture_num: unsupported texture type (%d)",
		      (int)tt);
   };
   return 0;
};
Texture *get_texture_bynum(TextureType tt,int num) {
   switch(tt) {
   case(TT_FLAT): return flats+num;
   case(TT_WALL): return walltexs+num;
   default: logprintf(LOG_ERROR,'T',
		      "get_texture_bynum: unsupported texture type (%d)",
		      (int)tt);
   };
   return NULL;
};
int count_textures(TextureType tt) {
   switch(tt) {
   case(TT_FLAT): return nflats;
   case(TT_WALL): return nwalltexs;
   default: logprintf(LOG_ERROR,'T',
		      "count_textures: unsupported texture type (%d)",
		      (int)tt);
   };
   return 0;
};

void init_textures(void) {
   init_playpal();
   init_flats();
   init_sprites();
   init_walltexs();
#ifdef TXDEBUG
   logprintf(LOG_DEBUG,'T',"init_textures(): done");
#endif
};

void reset_textures(void) {
   reset_flats();
   reset_sprites();
   reset_walltexs();
   reset_patches();
   reset_fonts();
   reset_misc();
   reset_texture_lumps();
   reset_pnames();
   reset_playpal();
}

void free_texels(Texture *t) {
   if(t->texels==NULL) return;
   switch(t->type) {
   case(TT_FLAT): free_flat(t); break;
   case(TT_WALL): free_wtex(t); break;
   case(TT_MISC):
   case(TT_FONT):
   case(TT_SPRITE): free_sprite(t); break;
    default:
      logprintf(LOG_FATAL,'T',"Strange texture type (0x%x) in free_texels",t->type);
   };
};

void load_texels(Texture *t,int bpp) {
   /* some error checking */
   if(bpp!=1&&bpp!=2&&bpp!=4) 
     logprintf(LOG_FATAL,'T',"Strange BPP value (%d) in load_texels",bpp);
   if(t->texels&&t->bpp==bpp) return;
   if(t->texels) free_texels(t);
   /* now go do it */
   switch(t->type) {
   case(TT_FLAT): load_flat(t,bpp); break;
   case(TT_WALL): load_wtex(t,bpp); break;
   case(TT_MISC):
   case(TT_FONT):
   case(TT_SPRITE): load_sprite(t,bpp); break;
    default:
      logprintf(LOG_FATAL,'T',"Strange texture type (0x%x) in load_texels",t->type);
   };
   /* find average texel */
   switch(bpp) {
    case(1):t->avg_texel=*(unsigned char *)(t->texels);break;
    case(2):t->avg_texel=*(unsigned short *)(t->texels);break;
    case(4):t->avg_texel=*(unsigned int *)(t->texels);break;
   }
};
